from __future__ import annotations
from abc import ABC, abstractmethod

class Parameter(ABC):
    """ Base class for a parameter in a Markov Chain Monte Carlo simulation.
    """

    def __init__(self, value) -> None:
        """ Initialize the Parameter.

        :param value: The initial value of the parameter.
        """
        self.value = value

    @abstractmethod
    def update(self, value, accepted=True): 
        """ Update the value of the parameter, if accepted.

        value: The new value of the parameter.
        accepted: Whether the new value is accepted. Default true.
        """
        ...

    @abstractmethod
    def propose(self, key) -> Parameter: 
        """ Propose a new parameter given a key

        :param key: the key to generate the parameter with
        :return: A new parameter proposal
        """
        ...

    @abstractmethod
    def log_prior(self):
        """ Logarithm of the prior probability of the parameter.

        :return: The logarithm of the prior probability of the parameter.
        """
        ...




from typing import Type

from .tree import TopologyNode, HypTree

def symmetric_topology(height : int, degree : int, fake_root : bool = False) -> TopologyNode:
    """ Generate tree of given height and degree

    A tree of height zero contains just the root;
    a tree of height one contains the root and one level of leaves below it,
    and so forth.

    :param height: The height of the tree
    :param degree: The degree of each node in the tree
    :param new_node: The node used to construct the tree, defaults to TreeNode
    :param fake_root: The fake root node, defaults to None
    :raises ValueError: If height is negative
    :return: The constructed tree
    """
    assert height >= 0, f'Height expected to be nonngeative, received {height=}.'

    def _builder(h: int, degree: int, parent):
        node = TopologyNode(); node.parent = parent; node.children = []
        if h > 1:
            node.children = [_builder(h - 1, degree, node) for _ in range(degree)]
        return node

    return _builder(height + 1, degree, None)



def asymmetric_topology(h: int)  -> TopologyNode:
    """ 
    Generate an asymmetric binary tree of given height.
    A tree of height zero contains just the root;
    a tree of height one contains the root and one level of leaves below it, and so forth.

    :param h: The height of the tree
    :raises ValueError: If height is negative
    :return: The constructed tree
    """
    if h < 0:
        raise ValueError(f'Height shall be nonnegative integer, received {h=}.')
    elif h == 0:
        return ChildList([TreeNode()])

    # Fake root 
    root = TopologyNode(); root.parent = None; root.children = []
    root.children = [TopologyNode(children= [],parent=root), TopologyNode(children=[],parent=root)]
   
    node = root.children[0]

    for _ in range(h - 1):
        node.children=  [TopologyNode(children=[],parent=node), TopologyNode(children=[],parent=node)]

        node = node.children[0]

    return root 


### Alternative Newick tree generation
def read_topology(newick_str: str,return_topology=False) -> HypTree:
    """ 
    Generate a tree from a Newick string recursively.

    :param newick_str: newick string representation
    :return: The constructed tree
    """
    import re

    iter_tokens = re.finditer(r"([^:;,()\s]*)(?:\s*:\s*([\d.]+)\s*)?([,);])|(\S)", newick_str+";")

    def recursive_parse_newick(parent=None):
        name, length, delim, char = next(iter_tokens).groups(0)

        node = TopologyNode(name=name if name else None,             # create a "ghost" subtree root node without data
                       
                       parent=parent,
                       children=[])
        node.data = {"edge_length": float(length)} if length else {}

        
        if char == "(": # start a subtree

            while char in "(,": # add all children within a parenthesis to the current node
                child, char = recursive_parse_newick(parent=node)
                node.children.append(child)

            name, length, delim, char = next(iter_tokens).groups(0)

            node.name = name                        # assign data to the "ghost" subtree root node
            node.data = {"edge_length": float(length)} if length else {}
            
        return node, delim
    
    root, _ = recursive_parse_newick()


    if return_topology:
        return root

    else: 
        tree = HypTree(root)

        if 'edge_length' in root.data.keys():
            tree.add_property('edge_length', (1,), dtype=float)

            for node in tree.iter_topology_bfs():
                tree.data['edge_length'] = tree.data['edge_length'].at[node.id].set(node.data['edge_length'])
                del node.data

        return tree
    

    
def write_topology(tree:HypTree) -> str:
    """
    Convert a tree to a Newick string representation.

    :param tree: The tree to convert
    :return: Newick string representation of the tree
    """
    # Check if edge_length exists, otherwise fill it with ones
    if "edge_length" not in tree.data:
        tree.add_property('edge_length', shape=(1,))
        tree.data['edge_length'] =  tree.data['edge_length'].at[:].set(1)

    def recursive_to_newick(node) -> str:
        edge_length = tree.data["edge_length"].at[node.id].get().item()

        if not node.children:
            return f"{node.name}:{edge_length:.4f}" if node.name else f":{edge_length:.4f}"
        
        children_str = ",".join(recursive_to_newick(child) for child in node.children)
        node_str = f"({children_str}){node.name}:{edge_length:.4f}" if node.name else f"({children_str}):{edge_length:.4f}"
        return node_str

    return recursive_to_newick(tree.topology_root) + ";"



from .basemodel import BaseModel, ReducerModel
from abc import abstractmethod
from inspect import getfullargspec
from .utils import filter_keywords

class UpReducer(ReducerModel):
    def __init__(self, reductions) -> None:
        super().__init__(reductions=reductions)

    @abstractmethod
    def transform(self, **kwargs): ...

    @abstractmethod
    def up(self, **kwargs): ...

    def _set_keys(self):
        up_arg_spec = getfullargspec(self.up)
        upkeys = filter_keywords(up_arg_spec.args)
        self.up_keys = [k for k in upkeys if k != 'self']
        self.up_keys

        transform_arg_spec = getfullargspec(self.transform)
        keys = filter_keywords(transform_arg_spec.args)

        transform_child_keys = [k.removeprefix('child_') for k in keys if k.startswith('child_')]
        transform_parent_keys = [k for k in keys if not k.startswith('child_')]

        self.transform_child_keys = transform_child_keys
        self.transform_parent_keys = transform_parent_keys

class DownModel(BaseModel):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def down(self, **kwargs): ...

    def _set_keys(self):
        arg_spec = getfullargspec(self.down)

        keys = filter_keywords(arg_spec.args)

        down_parent_keys = [k.removeprefix('parent_') for k in keys if k.startswith('parent_')]
        down_child_keys = [k for k in keys if not k.startswith('parent_')]

        self.down_parent_keys = down_parent_keys
        self.down_child_keys = down_child_keys

class UpModel(BaseModel):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def up(self, **kwargs): ...

    def _set_keys(self):
        arg_spec = getfullargspec(self.up)

        keys = filter_keywords(arg_spec.args)

        up_child_keys = [k.removeprefix('child_') for k in keys if k.startswith('child_')]
        up_current_keys = [k for k in keys if not k.startswith('child_')]

        self.up_child_keys = up_child_keys
        self.up_current_keys = up_current_keys


from .basemodel import ReducerModel, BaseModel
from abc import abstractmethod
from ..mcmc import ParameterStore
from inspect import getfullargspec
from .utils import filter_keywords

class UpdateReducer(ReducerModel):
    def __init__(self, reductions) -> None:
        super().__init__(reductions=reductions)

    @abstractmethod
    def update(self, **kwargs): ...

    @abstractmethod
    def up(self, **kwargs): ...

    def _set_keys(self):
        up_arg_spec = getfullargspec(self.model.up)
        upkeys = up_arg_spec.args
        filter_keywords(upkeys)
        self.up_keys = upkeys


        update_arg_spec = getfullargspec(self.model.update)
        keys = filter_keywords(update_arg_spec.args)

        update_parent_keys = [k.removeprefix('parent_') for k in keys if k.startswith('parent_')]
        update_node_keys = [k for k in keys if not k.startswith('parent_') and not k.startswith('child_')]

        self.update_parent_keys = update_parent_keys
        self.update_node_keys = update_node_keys

class UpdateModel(BaseModel):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def update(self, **kwargs): ...

    def _set_keys(self):
        update_arg_spec = getfullargspec(self.update)
        keys = filter_keywords(update_arg_spec.args)
        update_child_keys = [k.removeprefix('child_') for k in keys if k.startswith('child_')]
        update_parent_keys = [k.removeprefix('parent_') for k in keys if k.startswith('parent_')]
        update_current_keys = [k for k in keys if not k.startswith('parent_') and not k.startswith('child_')]

        self.update_child_keys = update_child_keys
        self.update_parent_keys = update_parent_keys
        self.update_current_keys = update_current_keys


from jax.random import PRNGKey
from hyperiax.tree.topology import symmetric_topology

def test_symmetric_tree():
    root = symmetric_topology(3, 100)

    assert len(root.children) == 100
    for c in root.children:
        assert len(c.children) == 100


from ..models import UpdateReducer, UpdateModel
import jax
from inspect import getfullargspec
from functools import partial
from jax import numpy as jnp

class UnorderedExecutor:
    "Executes nodes in an unstructured way; giving full context from both parents and children"

    def __init__(self, model, key = None) -> None:
        self.model = model

        self.key = key if key else jax.random.PRNGKey(0)
        self.model = model

    def update(self, tree, params = {}):
        if issubclass(type(self.model), UpdateReducer):
            new_data = self._update_reduce_inner(tree.data, tree, params)
        elif issubclass(type(self.model), UpdateModel):
            coloring = tree.coloring
            red_indx = jnp.arange(tree.size)[coloring]
            new_data = self._update_inner(tree.data, tree, red_indx, params)
            black_indx = jnp.arange(tree.size)[~coloring]
            new_data = self._update_inner(new_data, tree, black_indx, params)
        else:
            raise ValueError('Model needs to be of type UpdateReducer or UpdateModel')
        tree.data = {**tree.data, **new_data}

    @partial(jax.jit, static_argnames=['tree', 'self']) 
    def _update_inner(self, data, tree, coloring, params):
        node_data = {
            k: data[k][coloring]
            for k in self.model.update_current_keys
        }
        parent_data = {
            f'parent_{k}': data[k][tree.parents[coloring]]
            for k in self.model.update_parent_keys
        }
        child_data = {
            f'child_{k}': data[k][tree.gather_child_idx[coloring]]
            for k in self.model.update_child_keys
        }

        result = self.model.update(**parent_data, **node_data, **child_data, 
                        root_mask=tree.is_root[coloring], 
                        leaf_mask=tree.is_leaf[coloring], params = params)

        for k, val in result.items():
            data[k] = data[k].at[coloring].set(val)

        return data


    @partial(jax.jit, static_argnames=['tree', 'self']) 
    def _update_reduce_inner(self, data, tree, params):
        iterator = zip(
            reversed(tree.levels[1:]), # indices for levels
            reversed(tree.pbuckets_ref) # indices for upwards propagation
        )
        unrolled = list(iterator)
        i1 = unrolled[::2]
        i2 = unrolled[1::2]
        # we can make this much faster by actually merging the underlying structures... #TODO
        for it in (i1, i2):
            for (level_start, level_end), up_ref in it:
                child_node_data = {
                    k: jax.lax.slice_in_dim(data[k], level_start, level_end)
                    for k in self.up_keys
                }
                #print(child_node_data)
                up_result = self.model.up(**child_node_data, params = params)

                segments = tree.pbuckets[level_start:level_end]

                fuse_scatter = {
                    f'child_{k}': self.model.reductions[k](v, segments, num_segments=len(up_ref), indices_are_sorted=True)
                    for k,v in up_result.items()
                }

                node_data = {
                    k: data[k][up_ref]
                    for k in self.update_node_keys
                }

                parent_data = {
                    f'parent_{k}': data[k][tree.parents[up_ref]]
                    for k in self.update_parent_keys
                }

                result = self.model.update(**parent_data, **node_data, **fuse_scatter, params = params)
                for k, val in result.items():
                    data[k] = data[k].at[up_ref].set(val)


        return data


from ..tree import HypTree

def plot_tree_2d_scatter(tree : HypTree, property : str, ax=None):
    from matplotlib import pyplot as plt
    from matplotlib import patches as mpatch


    cmap = plt.cm.ocean

    if ax == None:
        fig,ax = plt.subplots(figsize=(10,8))

    for i in range(1, len(tree)):
        cdat = tree.data[property][i]
        if i == 0: continue
        parentidx = tree.parents[i]
        dat = tree.data[property][parentidx]
        ax.arrow(*dat, *(cdat-dat), width=0.01, length_includes_head=True, color='gray')

    for i, (l_start, l_end) in enumerate(tree.levels):
        for nodeidx in range(l_start, l_end):
            cdat = tree.data[property][nodeidx]
            ax.scatter(*cdat, color=cmap(i/len(tree.levels)))

        handles = [mpatch.Patch(color=cmap(i/len(tree.levels)), label = f'{i+1}') for i in range(len(tree.levels))]
        legend = ax.legend(handles=handles, title="Levels")
        ax.add_artist(legend)
        ax.grid(True)



import hyperiax
from jax.random import PRNGKey, split
from hyperiax.execution import LevelwiseTreeExecutor
from hyperiax.tree.fasttree import HypTree
import jax
from hyperiax.tree.builders import symmetric_topology
import time
from matplotlib import pyplot as plt
from tqdm.auto import tqdm
key = PRNGKey(0)

depths = list(range(1, 19, 1))
newt = []
oldt = []

for depth in tqdm(depths):

    topology = symmetric_topology(depth, 2)
    fast_tree = HypTree(topology)
    #print(f"Fast tree with {len(fast_tree)} nodes")

    fast_tree.add_property('noise', (10,2))


    start = time.time()
    for level_start, level_end in fast_tree.levels:
        # start by fusing children  
        data = jax.lax.slice_in_dim(fast_tree.data['noise'].data, level_start, level_end)
            

    end = time.time()
    #print(f"New execution time: {end-start}")
    newt.append(end-start)

    tree = hyperiax.tree.builders.symmetric_tree(depth,2)
    #print(f"Old tree with {len(tree)} nodes")
    subkey, key = split(key)
    noise_tree = hyperiax.tree.initializers.initialize_noise(tree, key, (10, 2))

    lwe = LevelwiseTreeExecutor(None, batch_size=5000)
    noise_tree.order = (lwe._determine_execution_order(noise_tree), None)
    order = reversed(noise_tree.order[0])

    start = time.time()
            

    end = time.time()
    #print(f"Old execution time: {end-start}")
    oldt.append(end-start)

plt.plot(depths, newt, label='New')
plt.plot(depths, oldt, label='Old')
plt.legend()
plt.xlabel('Depth')
plt.ylabel('Time (s)')
plt.yscale('log')
plt.savefig('batch_generation.png')



def pass_up(*args):
    """ Returns a function that passes a set of keys on.

    """
    def _f(**kwargs):
        """ _F to pass_up

        :return: `pass_up('value', 'noise')` will give a function that given a dictionary, returns a new one containing only `value` and `noise`
        """
        return {k: kwargs[k] for k in args}
    return _f


from abc import ABC, abstractmethod
from .utils import map_reduction

class BaseModel(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._set_keys()
    
    @abstractmethod
    def _set_keys(self): ...

class ReducerModel(BaseModel):
    def __init__(self, reductions) -> None:
        super().__init__()
        self.reductions = {k: map_reduction(v) for k,v in reductions.items()}


from ..tree import HypTree, TopologyNode

def plot_tree_text(tree):
    formatter = TreeFormatter(tree)
    formatter.print_tree()


# Functions for nicer tree printing in the terminal, modified from https://github.com/AharonSambol/PrettyPrintTree

from typing import Any, List, Tuple, Generator

def zip_longest(*iterables: List, default: Any) -> Generator:
    """
    Returns an iterator that aggregates elements from each of the input iterables.
    If the iterables are of unequal length, missing values are filled with the specified default value.

    Args:
        *iterables: Variable number of iterables to be zipped together.
        default: The value to be used as a default when an iterable is exhausted.

    Yields:
        A tuple containing elements from each of the input iterables.

    Raises:
        None.

    Examples:
        >>> list(zip_longest([1, 2, 3], [4, 5], [6, 7, 8, 9], default=None))
        [(1, 4, 6), (2, 5, 7), (3, None, 8), (None, None, 9)]
    """
    lens = [len(x) for x in iterables]
    for line in range(max(lens)):
        yield tuple(
            iterable[line] if line < lens[i] else default
            for i, iterable in enumerate(iterables)
        )

def left_adjust(text: str, amount: int, padding: str = ' ') -> str:
    """
    Adjusts the given text to the left by adding padding characters.

    Args:
        text (str): The text to be adjusted.
        amount (int): The desired width of the adjusted text.
        padding (str, optional): The character used for padding. Defaults to ' '.

    Returns:
        str: The adjusted text.

    """
    return text + padding * (amount - len(text))

def right_adjust(text: str, amount: int, padding: str =' ') -> str:
    """
    Adjusts the given text to the right by adding padding characters.

    Args:
        text (str): The text to be right adjusted.
        amount (int): The total width of the resulting string.
        padding (str, optional): The character used for padding. Defaults to ' '.

    Returns:
        str: The right adjusted text.

    """
    return padding * (amount - len(text)) + text

class TreeNodeFormatter:
    """
    A class that represents a formatter for tree nodes.
    """

    @classmethod
    def from_string(cls: Any, string: str) -> Any:
        """
        Create a TreeNodeFormatter object from a string representation.

        Args:
            string (str): The string representation of the tree node.

        Returns:
            TreeNodeFormatter: The TreeNodeFormatter object.
        """
        lines = string.split('\n')
        height = len(lines)
        width = max(len(line) for line in lines)
        return cls(lines, height=height, width=width)
    
    def __init__(self, lines: List[str], *, height: int, width = int, middle_height: int = None, middle_width: int = None) -> None:
        """
        Initialize a TreeNodeFormatter object.

        Args:
            lines (list[str]): The content of the tree node.
            height (int): The height of the tree node text.
            width (int): The width of the tree node text.
            middle_height (int, optional): The middle height position of the tree node text. Defaults to None.
            middle_width (int, optional): The middle width position of the tree node text. Defaults to None.
        """
        self.lines = lines
        self.height = height
        self.width = width
        self._middle_height = middle_height
        self._middle_width = middle_width

    def to_string(self) -> str:
        """
        Convert the TreeNodeFormatter object to a string representation.

        Returns:
            str: The string representation of the tree node.
        """
        return '\n'.join(self.lines)
    
    @property
    def middle_width(self) -> int:
        """
        Get the middle width of the tree node text.

        Returns:
            int: The middle width position of the tree node text.
        """
        if self._middle_width is None:
            return sum(divmod(self.width, 2)) - 1
        return self._middle_width
    
    @property
    def middle_height(self) -> int:
        """
        Get the middle height of the tree node text.

        Returns:
            int: The middle height position of the tree node text.
        """
        if self._middle_height is None:
            return sum(divmod(self.height, 2)) - 1
        return self._middle_height

def add_parent(parent_formatter: TreeNodeFormatter, children_formatter: TreeNodeFormatter) -> TreeNodeFormatter:
    """
    Combines the formatting information of a parent node and its children nodes to create a new TreeNodeFormatter object.

    Args:
        parent_formatter (TreeNodeFormatter): The formatting information of the parent node.
        children_formatter (TreeNodeFormatter): The formatting information of the children nodes.

    Returns:
        TreeNodeFormatter: A new TreeNodeFormatter object with the combined formatting information.

    """
    parent_middle, children_middle = parent_formatter.middle_width, children_formatter.middle_width
    parent_width, children_width = parent_formatter.width, children_formatter.width
    
    if parent_middle == children_middle:      # Only one child
        lines = parent_formatter.lines + children_formatter.lines
        middle = parent_middle
    elif parent_middle < children_middle:     # Have several children
        padding = ' ' * (children_middle - parent_middle)
        lines = [padding + line for line in parent_formatter.lines] + children_formatter.lines
        parent_width += children_middle - parent_middle
        middle = children_middle
    else:                                    # Have no child
        padding = ' ' * (parent_middle - children_middle)
        lines = parent_formatter.lines + [padding + line for line in children_formatter.lines]
        children_width += parent_middle - children_middle
        middle = parent_middle
    
    return TreeNodeFormatter(
        lines,
        height=parent_formatter.height + children_formatter.height,
        width=max(parent_width, children_width),
        middle_width=middle
    )

class TreeFormatter:
    """
    A class that formats a HypTree object into a string representation.

    Attributes:
        root (TreeNode): The root node of the HypTree object.

    Methods:
        get_children(node: TreeNode) -> list[TreeNode]:
            Returns the children of a given node.
        
        get_name(node: TreeNode) -> str:
            Returns the name of a given node. If the name is None, returns '*'.
        
        add_name(name: str, node_formatter: TreeNodeFormatter, parent_adder: callable=add_parent, seperator: str = ' ') -> TreeNodeFormatter:
            Adds a name to the node formatter and returns the updated formatter.
        
        format() -> str:
            Formats the HypTree object into a string representation and returns it.
        
        tree_join_formatter(node: TreeNode, depth: int = 0) -> TreeNodeFormatter:
            Recursively formats the tree structure and returns the root node formatter.
        
        node_to_formatter(node: TreeNode) -> TreeNodeFormatter:
            Converts a node into a TreeNodeFormatter object and returns it.
        
        print_tree() -> None:
            Prints the formatted tree representation to the console.
    """

    def __init__(self, tree) -> None:
        if issubclass(type(tree), HypTree):
            self.root = tree.topology_root
        elif issubclass(type(tree), TopologyNode):
            self.root = tree
        else:
            raise ValueError('Tree must be of type HypTree or TopologyNode')

    def get_children(self, node: TopologyNode) -> List[TopologyNode]:
        return node.children
    
    def get_name(self, node: TopologyNode) -> str:
        return node.name if 'name' in node.__dict__ else '*'
    
    # def add_name(self, name: str, node_formatter: TreeNodeFormatter, parent_adder: callable=add_parent, seperator: str = ' ') -> TreeNodeFormatter:
    #     if name:
    #         name_formatter = TreeNodeFormatter.from_string(str(name))
    #         node_formatter = parent_adder(TreeNodeFormatter.from_string(seperator), node_formatter)
    #         node_formatter = parent_adder(name_formatter, node_formatter)
    #     return node_formatter
    
    def format(self) -> str:
        """
        Formats the tree structure into a string representation.

        Returns:
            str: The formatted tree structure as a string.
        """
        res = self.tree_join_formatter(self.root)
        res = res.to_string().rstrip()
        return res
    
    def tree_join_formatter(self, node: TopologyNode, depth: int = 0) -> TreeNodeFormatter:
        """
        Recursively formats the tree structure and returns the root node formatter.

        Args:
            node (TreeNode): The current node being formatted.
            depth (int): The depth of the current node in the tree structure.

        Returns:
            TreeNodeFormatter: The root node formatter.
        """
        # name = self.get_name(node)
        children = self.get_children(node)
        node_formatter = self.node_to_formatter(node)

        if children:
            children_formatters = [
                self.tree_join_formatter(child, depth=depth+1)
                for child in children
            ]
            if len(children) == 1:
                children_node_formatter = children_formatters[0]
                children_node_formatter.lines.insert(0, ' ' * children_node_formatter.middle_width + '│')
            else:
                children_node_formatter = join_horizontally(children_formatters)
            
            node_formatter = add_parent(node_formatter, children_node_formatter)
        
        # node_formatter = self.add_name(name, node_formatter)
            
        return node_formatter
    
    def node_to_formatter(self, node: TopologyNode) -> TreeNodeFormatter:
        """
        Converts a node into a TreeNodeFormatter object and returns it.

        Args:
            node (TreeNode): The node to convert.

        Returns:
            TreeNodeFormatter: The converted node formatter.
        """
        name = self.get_name(node)
        return TreeNodeFormatter.from_string(str(name))
    
    def print_tree(self) -> None:
        """
        Prints the formatted tree representation to the console.
        """
        print(self.format())

def join_horizontally(boxes: List[TreeNodeFormatter]) -> TreeNodeFormatter:
    """
    Joins multiple TreeNodeFormatter boxes horizontally to form a line to print.

    Args:
        boxes (list[TreeNodeFormatter]): A list of TreeNodeFormatter objects to gather.

    Returns:
        TreeNodeFormatter: The joined TreeNodeFormatter box.

    """
    lines, width, height = join_boxes(boxes)
    middle = add_pipes(boxes, lines)
    height += 1
    return TreeNodeFormatter(lines, height=height, width=width, middle_width=middle)

def join_boxes(boxes: List[TreeNodeFormatter]) -> Tuple[List[str], int, int]:
    """
    Joins multiple TreeNodeFormatter together horizontally to get the text, width and height of this line.

    Args:
        boxes (List[TreeNodeFormatter]): A list of TreeNodeFormatter objects to gather.

    Returns:
        Tuple[List[str], int, int]: A tuple containing the joined lines, width, and height of the joined boxes.
    """
    lines = [
        ' '.join(
            left_adjust(text=line, amount=boxes[i].width)
            for i, line in enumerate(lines)
        )
        for lines in zip_longest(*(box.lines for box in boxes), default='')
    ]
    width = sum(box.width for box in boxes) + len(boxes) - 1
    height = max(box.height for box in boxes)
    return lines, width, height

def add_pipes(boxes: List[TreeNodeFormatter], lines: List[str]) -> int:
    """
    Adds pipes to the given lines to create a tree-like structure.

    Args:
        boxes (List[TreeNodeFormatter]): A list of TreeNodeFormatter objects to gather.
        lines (List[str]): A list of strings representing the lines of the tree.

    Returns:
        int: The number of characters added as padding to the lines.

    """
    padding = ' ' * boxes[0].middle_width
    pipes = '┌'
    for prev, box in zip(boxes, boxes[1:]):
        pipes += '─' * (prev.width - prev.middle_width + box.middle_width) + '┬'
    middle_of_pipes = sum(divmod(len(pipes), 2)) - 1
    pipes = (
        padding 
        + pipes[:middle_of_pipes] 
        + {"─": "┴", "┬": "┼", "┌": "├", "┐": "┤"}[pipes[middle_of_pipes]]
        + pipes[middle_of_pipes + 1:-1]
        + '┐'
    )
    lines.insert(0, pipes)
    return len(padding) + middle_of_pipes


import numpy as np
from matplotlib import pyplot as plt

# trace plots for MCMC
def trace_plots(samples):
    """ Trace plots 
    
    :param samples: A list of ParameterStore objects.
    """

    # Determine grid layout
    num_params = len(samples[0])
    num_rows = int(np.sqrt(num_params))
    num_cols = num_params // num_rows + (num_params % num_rows > 0)

    # Create subplots
    fig, axes = plt.subplots(nrows=num_rows, ncols=num_cols, figsize=(15, 5))
    axes = np.atleast_2d(axes)  # Ensure axes is always 2D array

    # Plot traces
    for ax, (i, param) in zip(axes.ravel(), enumerate(samples[0].values().keys())):
        ax.plot(np.array([sample[param].value for sample in samples]))
        ax.set_title(f"Trace for {param}")
        ax.set_xlabel('Iteration')
        ax.set_ylabel(param)
        ax.grid(True)

    # Remove any unused subplots
    for ax in axes.ravel()[num_params:]:
        ax.axis('off')

    plt.tight_layout()
    plt.show()


from __future__ import annotations
from typing import Dict
from .parameter import Parameter
from jax.random import split

class ParameterStore:
    """ The parameter store hosts a set of parameters. 
    """

    def __init__(self, params : Dict[str, Parameter]) -> None:
        """ Initialize the parameter store.

        :param params: Dict with the parameters to store and their associated keys (names).
        """
        self.params = params

    def __len__(self):
        """ Get the number of parameters in the store.

        :return: The number of parameters in the store
        """
        return len(self.params)

    def __getitem__(self, key):
        return self.params[key]

    def values(self):
        """ Gets the values of the parameters in the store

        :return: The parameters in a dictionary.
        """
        return {k: v.value for k,v in self.params.items()}


    def propose(self, key) -> ParameterStore:
        """ Propose a new parameter store, containing newly sampled parameters.

        :param key: A key to sample with.
        :return: The newly sampled parameter store.
        """
        pcount = len(self)
        subkeys = split(key, pcount)
        
        return ParameterStore({
            k: v.propose(rngkey)
            for 
            rngkey, (k,v)
            in 
            zip(subkeys, self.params.items())
        })

    def log_prior(self):
        """ Logarithm of the prior probability of the parameters.

        :return: The logarithm of the prior probability of the parameters.
        """
        return sum([v.log_prior() for v in self.params.values()])




from .parameter import Parameter
from copy import copy
class FixedParameter(Parameter):
    """ A parameter that is constant regardless of updates.  The parameter always have the value passed in the constructor
    """
    def __init__(self, value) -> None:
        super().__init__(value)

    def propose(self, key):
        """ Propose a new value for the parameter

        :param key: A key to sample with
        :return: Copy of the parameter
        """
        return copy(self)
    
    def update(self, value, accepted): 
        """ Update the parameter. Does nothing as the parameter is fixed, i.e. the new value is discarded

        :param value: The new value to update the parameter with
        :param accepted: Whether the new value was accepted
        """
        ...

    def log_prior(self):
        """ Logarithm of the prior probability of the parameter

        :return: 0. as the parameter is fixed
        """
        return 0.




from jax import ops

def map_reduction(reduction_string):
    map = {
        'sum': ops.segment_sum,
        'prod': ops.segment_prod,
        'max': ops.segment_max,
        'min': ops.segment_min
    }
    if reduction_string not in map.keys():
        raise ValueError(f"Reduction '{reduction_string}' not supported")
    return map[reduction_string]


def filter_keywords(keys):
    reserved_keywords = ['self', 'params', 'key', 'root_mask', 'leaf_mask']

    return [k for k in keys if k not in reserved_keywords]



from . import HypTree, TreeNode
from jax.random import PRNGKey, split
import jax
import jax.numpy as jnp
import copy
from typing import Callable
from functools import partial

def update_noise_inplace(update_f: Callable[[TreeNode,float],float], tree : HypTree, key=False, save_noise=False) -> HypTree:
    raise DeprecationWarning("This function is deprecated")
    for node in tree.iter_bfs():
        subkey, key = split(key)
        if save_noise:
            node.data['old_noise'] = node.data['noise']
        node.data['noise'] = update_f(node,jax.random.normal(subkey, node.data['noise'].shape))

    return tree




def sum_fuse_children(axis=-1):
    """ Fuse across children 

    :param axis:  axis (int, optional): the axis to sum over. Defaults to -1.
    """
    def _fuse(**kwargs):
        """ _fuse

        :return: returns a function that sums an axis of keyed values prefixed with `child_`
        """
        return {k[6:]:v.sum(axis=axis) for k,v in kwargs.items() if k.startswith('child_')}
    return _fuse




import jax
import jax.numpy as jnp
import numpy as np
from tqdm import tqdm

def metropolis_hastings(log_target, proposal, data, init, num_samples, burn_in=0, thin=1, rng_key=None, savef=lambda x: x):
    """ MCMC using Metropolis-Hastings algorithm

    :param log_target: Function returning the logarithm of the target distribution
    :param proposal: Function returning a sample from the proposal distribution
    :param data: Observed data
    :param init: Initial state
    :param num_samples: Number of samples to draw
    :param burn_in: Number of samples to discard at the beginning
    :param thin: Thinning rate
    :param rng_key: Random number generator key
    :param savef: Function to save the state
    :return: List of log likelihoods and samples
    """
    if rng_key is None:
        rng_key = jax.random.PRNGKey(0)

    current_state = init
    log_likelihood_current = log_target(data,current_state)
    log_likelihoods = []
    samples = []
    accepted_count = 0  # Initialize counter for accepted proposals

    for i in tqdm(range(num_samples + burn_in)):
        rng_key, subkey = jax.random.split(rng_key)

        # Propose a new state
        proposed_state = proposal(current_state, subkey)

        # Compute acceptance probability
        log_likelihood_proposed = log_target(data,proposed_state)
        alpha = jnp.minimum(1.0, jnp.exp(log_likelihood_proposed - log_likelihood_current))
        #print(alpha,log_likelihood_proposed,log_likelihood_current,proposed_state[0].values(),current_state[0].values())

        # Decide whether to accept the proposed state
        if jax.random.uniform(subkey) < alpha:
            current_state = proposed_state
            log_likelihood_current = log_likelihood_proposed
            accepted_count += 1  # Increment count if proposal is accepted

        # Save the sample after burn-in and according to thinning rate
        if i >= burn_in and (i - burn_in) % thin == 0:
            log_likelihoods.append(log_likelihood_current)
            samples.append(savef(current_state))

    # acceptance rate
    acceptance_rate = accepted_count / (num_samples + burn_in)
    print(f"Acceptance rate: {acceptance_rate:.4f}")

    return log_likelihoods,samples


from .parameter import Parameter

import jax
import jax.numpy as jnp
from jax.scipy.special import gammaln

def inverse_gamma_logpdf(x, alpha, beta):
    """ Log PDF of the inverse gamma distribution
    
    :param x: The value to evaluate the PDF at
    :param alpha: The alpha parameter of the inverse gamma distribution
    :param beta: The beta parameter of the inverse gamma distribution
    :return: The logarithm of the PDF of the inverse gamma distribution at x
    """
    return alpha * jnp.log(beta) - gammaln(alpha) - (alpha + 1) * jnp.log(x) - beta / x

class VarianceParameter(Parameter):
    def __init__(self, value, alpha=3., beta=2., proposal_var=.01, keep_constant=False) -> None:
        """ Initialize a variance parameter
        
        :param value: The initial value of the parameter
        :param alpha: The alpha parameter of the inverse gamma distribution
        :param beta: The beta parameter of the inverse gamma distribution
        :param proposal_var: The variance of the proposal distribution
        :param keep_constant: Whether the parameter should be kept constant in new proposals
        """
        super().__init__(value)
        self.alpha = alpha
        self.beta = beta
        self.proposal_var = proposal_var
        self.keep_constant = keep_constant

    def propose(self, key):
        """ Propose a new value for the parameter
        
        :param key: A key to sample with
        :return: The newly proposed parameter
        """
        if self.keep_constant:
            return self

        return VarianceParameter(jnp.exp(jnp.log(self.value)+jnp.sqrt(self.proposal_var)*jax.random.normal(key)), self.alpha, self.beta, self.proposal_var, self.keep_constant)
    
    def update(self, value, accepted): 
        """ Update the parameter with a new value
        
        :param value: The new value to update the parameter with
        :param accepted: Whether the new value was accepted.
        """
        if accepted:
            self.value = value
    
    def log_prior(self):
        """ Logarithm of the prior probability of the parameter
        
        :return: The logarithm of the prior probability of the parameter
        """
        return inverse_gamma_logpdf(self.value,self.alpha,self.beta)



# Functions for plotting data and tree


#####################################################################################################
# 2d plot of data points in a 

def plot_tree_2d_(self, ax=None, selector=None):
    from matplotlib import pyplot as plt
    from matplotlib import patches as mpatch

    tree = 'partial'
    for node in self.iter_bfs():
        if node.data == None: break
    else:
        tree = 'full'

    cmap = plt.cm.ocean

    if ax == None:
        fig,ax = plt.subplots(figsize=(10,8))
    if tree == 'full':
        levels = list(self.iter_levels())

        for i, level in enumerate(levels):
            for node in level:
                dat = selector(node.data) if selector else node.data
                if node.children:
                    for child in node.children:
                        cdat = selector(child.data) if selector else child.data
                        ax.arrow(*dat, *(cdat-dat), width=0.01, length_includes_head=True, color='gray')
                ax.scatter(*dat, color=cmap(i/len(levels)))
                if 'name' in node.data.keys():
                    ax.annotate(node.data['name'], dat, xytext=(5,5), textcoords='offset pixels')

        handles = [mpatch.Patch(color=cmap(i/len(levels)), label = f'{i+1}') for i in range(len(levels))]
        legend = ax.legend(handles=handles, title="Levels")
        ax.add_artist(legend)
        ax.grid(True)

#####################################################################################################
# tree illustration 
        
def estimate_position(self):
    """ Estimate the x and y coordinates of each point """
    
    # Determine Y coordinates, with distance from root 
    self.root.data["y_temp"] = 0
    for leaf in self.iter_dfs():
        # initlize an empty x coordinate for later 
        leaf.data["x_temp"] = 0
        if leaf.parent is not None: # Skip root
            if 'edge_length' in leaf.data.keys():
                leaf.data["y_temp"] = leaf.parent.data["y_temp"] -leaf.data["edge_length"] 

            else: 
                leaf.data["y_temp"] = leaf.parent.data["y_temp"] - 1
    # Define x coordinate for each leaf 
    for i,leaf in enumerate(self.iter_leaves_dfs()):
        leaf.data["x_temp"] = i


    # Determine X coordinates from bottom and up 
    for level in reversed(list(self.iter_levels())):
        for leaf in level:
                while leaf.parent is not None:
                    x_coordinate = 0
                    leaf = leaf.parent
                    for i,node in enumerate(leaf.children):
                        x_coordinate += node.data["x_temp"]
                    leaf.data["x_temp"] = x_coordinate/(i+1)
    return self 
     
def plot_tree_(self,ax=None,inc_names=False): 
    from matplotlib import pyplot as plt
    """Plot the tree using matplotlib"""
 
    if ax == None:
        fig,ax = plt.subplots(figsize=(10,8))

    self = estimate_position(self)
    ax.plot(self.root.data["x_temp"], self.root.data["y_temp"], 'ko')  # Plot the current node
    ax.axis('off')
    for leaf in self.iter_bfs():
        if len(leaf.children) != 0:
            plot_node(leaf,ax,inc_names)

def plot_node(parent,ax,inc_names):
        from matplotlib import pyplot as plt
        """Plot a single node and its children"""
        ax.plot(parent.data["x_temp"], parent.data["y_temp"], 'ko')  # Plot the current node
        if inc_names and parent.name is not None:
            ax.text(parent.data["x_temp"], parent.data["y_temp"], parent.name+" ", fontdict=None,rotation="vertical",va="top",ha="center")
            # Draw vertical line from parent to current level
        for child in parent.children:
            ax.plot(child.data["x_temp"], child.data["y_temp"], 'ko')

            # Include text 
            if inc_names and child.name is not None:
                ax.text(child.data["x_temp"], child.data["y_temp"], child.name+" ", fontdict=None,rotation="vertical",va="top",ha="center")
            # Draw vertical line from parent to current level
            ax.plot([child.data["x_temp"], parent.data["x_temp"]], [parent.data["y_temp"], parent.data["y_temp"]], 'k-')
            # Draw horizontal line to child
            ax.plot([child.data["x_temp"], child.data["x_temp"]], [parent.data["y_temp"], child.data["y_temp"]], 'k-') 





# Configuration file for the Sphinx documentation builder.
import os
import sys
sys.path.insert(0, os.path.abspath('../../hyperiax/'))

# -- Project information

project = 'Hyperiax'
copyright = '2024, CCEM, UCPH'
author = 'CCEM'

release = '2.0'
version = '2.0'

# -- General configuration

extensions = [
    'nbsphinx',
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx_autodoc_annotation',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
    'jax': ('https://jax.readthedocs.io/en/latest/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

html_theme = 'sphinx_rtd_theme'

# -- Options for EPUB output
epub_show_urls = 'footnote'

nbsphinx_allow_errors = True
nbsphinx_execute = 'never'



from hyperiax.models.functional import pass_up, sum_fuse_children
from hyperiax.tree import TopologyNode, HypTree
from jax import numpy as jnp

from hyperiax.models.lambdamodels import UpLambdaReducer
from hyperiax.execution import OrderedExecutor
from hyperiax.plotting import plot_tree_text
from hyperiax.models.lambdamodels import UpLambda, UpLambdaReducer
from test_fixtures import noise_tree


def test_asym():
    root = TopologyNode()
    root.children = [TopologyNode(root) for _ in range(4)]
    root.children[1].children = [TopologyNode(root.children[1]) for _ in range(2)]
    root.children[3].children = [TopologyNode(root.children[3])]
    tree = HypTree(root)
    tree.add_property('count', (1,))
    tree.add_property('sum_count', (1,))
    tree.data['count'] = jnp.ones_like(tree.data['count'])
    tree.data['sum_count'] = tree.data['sum_count'].at[tree.is_leaf].set(tree.data['count'][tree.is_leaf])
    def _up(sum_count, **kwargs):
        return {'sub_count': sum_count}
    def _transform(child_sub_count, count, **kwargs):
        return {'sum_count': child_sub_count+count}
    
    print()
    plot_tree_text(tree)
    
    upmodel = UpLambdaReducer(_up, _transform, {'sub_count': 'sum'})
    exe = OrderedExecutor(upmodel)
    exe.up(tree)

    assert (tree.data['sum_count'].squeeze() == jnp.array([8,1,3,1,2,1,1,1])).all()
    assert tree.data['sum_count'][0] == len(tree)

def test_lwte_sum(noise_tree):
    assert len(noise_tree.topology_root.children) == 2
    up = lambda noise, **kwargs: {'noise':noise}
    def transform(child_noise, noise,**kwargs): 
        return {'noise': noise+child_noise}

    model = UpLambdaReducer(up, transform_fn=transform, reductions={'noise': 'sum'})

    exe = OrderedExecutor(model)

    manual_sum = noise_tree.data['noise'].sum(axis=0)
    exe.up(noise_tree)

    assert jnp.isclose(noise_tree.data['noise'][0], manual_sum).all()

    #print




from .updownmodel import UpReducer,DownModel,UpModel
from .lambdamodels import UpLambdaReducer, DownLambda, UpdateLambda, UpLambda
from .updatemodel import UpdateReducer, UpdateModel
from . import functional    


from __future__ import annotations
from abc import ABC
from ..tree import HypTree
import jax
from ..models import UpReducer, DownModel, UpModel
from inspect import getfullargspec
from functools import partial

class OrderedExecutor(ABC):
    """Abstract for the ordered executor.

    Ordered executors performs operations in a certain direction (eg. up/down).

    It contains on abstract method `_determine_execution_order``, which gives the order in which nodes
    are to be executed in the tree. It is assumed that all operations are batched.
    """
    def __init__(self, 
                 model : UpReducer | DownModel,
                 key = None) -> None:
        self.key = key if key else jax.random.PRNGKey(0)
        self.model = model

    def up(self, tree : HypTree, params = {}):
        if not issubclass(type(self.model), (UpReducer, UpModel)):
            raise ValueError('Model needs to be of type UpReducer or UpModel')
        # out here we fetch the data stored in the tree

        #for k,val_shape in self.model.up_produces.items():
        #    if k not in tree.data:
        #        tree.add_property(k, val_shape)

        if issubclass(type(self.model), UpReducer):
            new_data = self._reduce_inner(tree.data, tree, params)

            tree.data = {**tree.data, **new_data}

            return
        elif issubclass(type(self.model), UpModel):
            new_data = self._up_inner(tree.data, tree, params = params)

            tree.data = {**tree.data, **new_data}

            return

    @partial(jax.jit, static_argnames=['tree', 'self']) 
    def _reduce_inner(self, data, tree, params):
        iterator = zip(
            reversed(tree.levels[1:]), # indices for levels
            #reversed(tree.psizes[:-1]), # indices for grouping calculations
            reversed(tree.pbuckets_ref) # indices for upwards propagation
        )
        for (level_start, level_end), up_ref in iterator:
            node_data = {
                k: jax.lax.slice_in_dim(data[k], level_start, level_end)
                for k in self.model.up_keys
            }
            #print(node_data)
            up_result = self.model.up(**node_data, params = params)

            segments = tree.pbuckets[level_start:level_end]

            fuse_scatter = {
                f'child_{k}': self.model.reductions[k](v, segments, num_segments=len(up_ref), indices_are_sorted=True)
                for k,v in up_result.items()
            }

            parent_data = {
                k: data[k][up_ref]
                for k in self.model.transform_parent_keys
            }

            result = self.model.transform(**parent_data, **fuse_scatter, params = params)
            for k, val in result.items():
                data[k] = data[k].at[up_ref].set(val)


        return data
    @partial(jax.jit, static_argnames=['tree', 'self']) 
    def _up_inner(self, data, tree, params):
        iterator = zip(
            reversed(tree.levels), # indices for levels
            #reversed(tree.psizes[:-1]), # indices for grouping calculations
            reversed(tree.pbuckets_ref) # indices for upwards propagation
        )
        next(iterator)
        for (level_start, level_end), up_ref in iterator:
            node_data = {
                k: data[k][level_start:level_end]
                for k in self.model.up_current_keys
            }

            child_data = {
                f'child_{k}': data[k][tree.gather_child_idx[level_start:level_end]]
                for k in self.model.up_child_keys
            }

            #print("CHILDDD", tree.gather_child_idx[level_start:level_end])

            result = self.model.up(
                **node_data,
                **child_data, 
                root_mask=tree.is_root[level_start:level_end], 
                leaf_mask=tree.is_leaf[level_start:level_end],
                params = params)
            
            for k, val in result.items():
                data[k] = data[k].at[level_start:level_end].set(val)

        return data



    @partial(jax.jit, static_argnames=['tree', 'self']) 
    def _down_inner(self, data, tree, params): #inner jitted runner for MAXIMUM PERFORMANCE
        #print(data, tree.levels)
        for level_start, level_end in tree.levels[1:]:
            node_data = {
                k: jax.lax.slice_in_dim(data[k], level_start, level_end)
                for k in self.model.down_child_keys
            }
            parent_indices = tree.parents[level_start:level_end] # no need to lax since tree is untraced
            parent_data = {
                f'parent_{k}': data[k][parent_indices]
                for k in self.model.down_parent_keys
            }

            if params:
                result = self.model.down(**node_data, **parent_data, params=params)
            else:
                result = self.model.down(**node_data, **parent_data)

            for k, val in result.items():
                data[k] = data[k].at[level_start:level_end].set(val)
        return data
    
    def down(self, tree : HypTree, params = {}):
        """Runs the down pass on the tree using the `down` function from the model.

        The `down` function in the model takes the following special parameters;

        `up_msg`: The message that was sent UP the edge during an upward pass
        `key`: A PRNGKey
        `params`: The same params passed to this function
        `parent_*`: Data from the parent node will be passed as arguments prefixed by `parent`.\
              If you have `value` stored in the parent node, `parent_value` will be an argument passed.
        `*`: Data from the child node will be passed as arguments without prefix. \
            If you have `value` stored in the child node, `value` will be an argument passed.

        The executor expects the `down` function to return a dictionary, \
            where each key returned overwrites the data in the lower node (with values not prefixed by `parent_`).
        Because of this, all values returned from `down` functions must have a leading axis of size `b`.

        Args:
            tree (HypTree): The tree to execute on
            params (dict, optional): optional parameters for mcmc. Defaults to {}.

        Returns:
            HypTree: The tree after running one pass
        """
        if not issubclass(type(self.model), DownModel):
            raise ValueError('Model needs to be of type DownModel')
        # out here we fetch the data stored in the tree

        new_data = self._down_inner(tree.data, tree, params = params)

        tree.data = {**tree.data, **new_data}

        return 


from ..tree import HypTree
from jax import numpy as jnp
import copy

        
def estimate_position(self):
    """ 
    Estimate the x and y position in the plotting coordinates of each node in the tree, and add the x and y position to the data dictionary of each node.

    :param tree: the tree to estimate the position in
    :return: the tree with the x and y position added to the data dictionary of each node
    """
    
    # Determine Y coordinates, with distance from.topology_root 
    depth = jnp.max(self.node_depths)
    self.data['y_temp'] = self.data['y_temp'].at[0].set(depth)
    for leaf in self.iter_topology_dfs():
        # initlize an empty x coordinate for later 
        self.data['x_temp'] = self.data['x_temp'].at[leaf.id].set(0)

        if leaf.parent is not None: # Skip.topology_root
            if hasattr(leaf, 'data') and 'edge_length' in leaf.data.keys():
           # if 'edge_length' in leaf.data.keys():
                new_value = self.data['y_temp'][leaf.parent.id]-leaf.data['edge_length']
                self.data['y_temp'] = self.data['y_temp'].at[leaf.id].set(new_value)

            else: 
                depth = jnp.max(self.node_depths)-self.node_depths[leaf.id]
                self.data['y_temp'] = self.data['y_temp'].at[leaf.id].set(depth)

    # Define x coordinate for each leaf 
    for i,leaf in enumerate(self.iter_topology_leaves_dfs()):
        self.data['x_temp'] = self.data['x_temp'].at[leaf.id].set(i)


    # Determine X coordinates from bottom and up 
    for level in reversed(list(self.iter_topology_levels())):
        for leaf in level:
                while leaf.parent is not None:
                    x_coordinate = 0
                    leaf = leaf.parent
                    for i,node in enumerate(leaf.children):
                        x_coordinate += self.data['x_temp'][node.id]
                    self.data['x_temp'] = self.data['x_temp'].at[leaf.id].set(x_coordinate/(i+1))

     
def plot_tree(self_input,ax=None,inc_names=False): 
    from matplotlib import pyplot as plt
    """
    Visualize the hierarchical structure of the tree

    :param tree: the tree to plot
    :param ax: the axis to plot the tree on, if None, a new figure is created, defaults to None
    :param inc_names: whether to include the names of the nodes in the plot, defaults to False
    """

    self = copy.deepcopy(self_input)

    self.add_property('x_temp', shape=(1,))
    self.add_property('y_temp', shape=(1,))
 
    if ax == None:
        fig,ax = plt.subplots(figsize=(10,8))


    estimate_position(self)

   
    ax.plot(self.data["x_temp"], self.data["y_temp"], 'ko') # plot all nodes
    ax.axis('off')
     
    # Plot the connection for each node 
    for leaf in self.iter_topology_bfs():
        if len(leaf.children) != 0:
           # 1+1
            plot_node(self,leaf,ax,inc_names)

def plot_node(self,parent,ax,inc_names):
    """
    Plot a single node and its children

    :param parent: the parent node to plot
    :param ax: the axis to plot the node on
    :param inc_names: whether to include the names of the nodes in the plot
    """
    from matplotlib import pyplot as plt
    """Plot a single node and its children"""
    ax.plot(self.data["x_temp"][parent.id],self.data["y_temp"][parent.id], 'ko')  # Plot the current node

    if inc_names:
        ax.text(self.data["x_temp"][parent.id].item(),self.data["y_temp"][parent.id].item(),  parent.name+"  ", fontdict=None,rotation="vertical",va="top",ha="center")
        1+1
    for child in parent.children:
        if inc_names:
            ax.text(self.data["x_temp"][child.id].item(),self.data["y_temp"][child.id].item(), child.name+"  ", fontdict=None,rotation="vertical",va="top",ha="center")
        
        
        # Draw vertical line to parent
        ax.plot([self.data["x_temp"][child.id],self.data["x_temp"][parent.id]], 
                [self.data["y_temp"][parent.id],self.data["y_temp"][parent.id]], 'k-')
        # Draw horizontal line to child
        ax.plot([self.data["x_temp"][child.id], self.data["x_temp"][child.id]], 
                [self.data["y_temp"][parent.id], self.data["y_temp"][child.id]], 'k-') 






from .updownmodel import UpReducer, DownModel, UpModel
from .updatemodel import UpdateReducer, UpdateModel

class UpLambdaReducer(UpReducer):
    """ Lambda model that only contains an up interface

    :param UpModel: Requires an up and fuse function.
    """
    def __init__(self, up_fn, transform_fn, reductions) -> None:
        """Lambda model that only contains an up interface

        :param up_fn: param intputs from node to transform_fn
        :param transform_fn: function to transform or do calculations form up parameters
        """

        self.up = up_fn
        self.transform = transform_fn

        super().__init__(reductions=reductions)

    def up(self, *args, **kwargs):
        """ Up function to define values to fuse function
        :return: input arguments to fuse function
        """
        raise ValueError('Model does not have a valid up function')
    
    def transform(self, *args, **kwargs):
        """ Fuse function to define calculations from up parameters
        :return: calculated values from up parameters to parent node
        """
        raise ValueError('Model does not have a valid fuse function')
    
class DownLambda(DownModel):
    """ Lambda model that only contains a down interface

    :param DownModel  Requires a down function.
    """

    def __init__(self, down_fn) -> None:
        """ Initialize DownLambda model

        :param down_fn: function to define values to fuse function
        """
        self.down = down_fn # need to do this to carry the argspec
    
        super().__init__()

    def down(self, *args, **kwargs):
        """ Down function to define values to fuse function
        :return: input arguments to fuse function
        """
        raise ValueError('Model does not have a valid down function')
    
class UpdateLambdaReducer(UpdateReducer):
    """Lambda model that only contains a local update interface

    :param UpdateModel: Requires an update function.
    """
    def __init__(self, up_fn, update_fn, reductions) -> None:
        """Lambda model that only contains an up interface

        :param up_fn: param intputs from node to update_fn
        :param update_fn: function to update or do calculations form up parameters
        """
        self.up = up_fn
        self.update = update_fn

        super().__init__(reductions=reductions)

    def up(self, *args, **kwargs):
        """ Up function to define values to fuse function
        :return: input arguments to fuse function
        """
        raise ValueError('Model does not have a valid up function')
    
    def update(self, *args, **kwargs):
        """ Fuse function to define calculations from up parameters
        :return: calculated values from up parameters to parent node
        """
        raise ValueError('Model does not have a valid fuse function')
    

class UpdateLambda(UpdateModel):
    """Lambda model that only contains a local update interface

    :param UpdateModel: Requires an update function.
    """
    def __init__(self, update_fn) -> None:
        """Lambda model that only contains an up interface

        :param update_fn: function to update or do calculations from both parent and child nodes
        """
        self.update = update_fn

        super().__init__()
    

    def update(self, *args, **kwargs):
        """ update function to define calculations from up parameters
        :return: calculated values from child and parent parameters
        """
        raise ValueError('Model does not have a valid update function')
    
class UpLambda(UpModel):
    """Lambda model that only contains an upward interface
    """
    def __init__(self, up_fn) -> None:
        """Lambda model that only contains an up interface

        :param up_fn: function to take current node and children node values and proive values to se in current
        """
        self.up = up_fn

        super().__init__()
    

    def up(self, *args, **kwargs):
        """ Up function to calculate the upward action on the tree
        :return: calculated values from child and current values
        """
        raise ValueError('Model does not have a valid up function')


import hyperiax
from jax.random import PRNGKey, split
import jax
from jax import numpy as jnp
from hyperiax.execution import LevelwiseTreeExecutor, DependencyTreeExecutor
from hyperiax.models import UpLambda, DownLambda
from hyperiax.models.functional import pass_up
import jax
from functools import partial
from hyperiax.models.prebuilts import PhyloMeanModel

import time

key = PRNGKey(0)

levels = 13

tree = hyperiax.tree.builders.symmetric_tree(levels,2)
print(f"Got tree with {len(tree)} nodes")
subkey, key = split(key)
noise_tree = hyperiax.tree.initializers.initialize_noise(tree, key, (2,))
### dummy pass_up function (for testing)

for i, node in enumerate(noise_tree.iter_bfs()):
    key, subkey = split(key)
    node['edge_length'] = jax.random.uniform(subkey, (1,))

@jax.jit
def down(noise, edge_length,parent_value, **args):
    return {'value': jnp.sqrt(edge_length)*noise+parent_value}

noise_tree.root['value'] = noise_tree.root['noise']

downmodel = DownLambda(down_fn=down)
exe = DependencyTreeExecutor(downmodel, batch_size=5000)
stoc_tree = exe.down(noise_tree)

for node in stoc_tree.iter_leaves():
    node['estimated_value'] = node['value']

pme = PhyloMeanModel()
exe = DependencyTreeExecutor(pme, batch_size=5000)

print("going")
prebuilt_tree = exe.up(stoc_tree)
print("done")


from ..updownmodel import UpReducer
import jax
from jax import numpy as jnp
from functools import partial
from ..functional import pass_up

class PhyloMeanModel(UpReducer):
    """Prebuilt model that calculates phylogenetic means in each node

    Assumes leaves all contain `estimated_value` before running it. This corresponds to actual observations.
    """
    def __init__(self, **kwargs) -> None:
        super().__init__(reductions={'weighted_value': 'sum', 'inverse_edge_length': 'sum'}, **kwargs)

    def up(self, estimated_value, edge_length, **kwargs):
        return {'weighted_value': estimated_value/edge_length, 'inverse_edge_length':1/edge_length}


    def transform(self, child_weighted_value,child_inverse_edge_length, **kwargs):
        return {'estimated_value': child_weighted_value/child_inverse_edge_length}


from __future__ import annotations
from collections import deque
from dataclasses import dataclass, field
from jax import numpy as jnp
from typing import Iterator, List
from functools import partial
import time
import jax

@dataclass
class TopologyNode:
    parent : TopologyNode = None
    children : List[TopologyNode] = field(default_factory=list)
    _backref : HypTree = None
    name : str = None
    def __getitem__(self, arg):
        if not self._backref:
            raise ValueError("This node is not part of a tree")
        return self._backref.data[arg][self.id]
    
class HypTree:
    def __init__(self, topology, precompute_child_gathers = False) -> None:
        self.topology_root = topology

        nodes = 0
        parents = [0]
        for i,node in enumerate(self.iter_topology_bfs()):
            nodes += 1
            node.id = i
            node._backref = self
            if node.parent: 
                parents.append(node.parent.id)
            #child_counts.append(len(node.children)) 

        
        
        
        self.size = nodes
        self.data = {}
        self.masks = {}


        self.levels = list(self._calculate_levels())
        self.node_depths = jnp.concatenate([i*jnp.ones((lend-lstart, 1), dtype=int) for i,(lstart,lend) in enumerate(self.levels)])

        self.parents = jnp.array(parents)
        self.child_counts = jax.ops.segment_sum(jnp.ones((self.size,), dtype=int), self.parents, num_segments=self.size, indices_are_sorted=True)
        self.child_counts = self.child_counts.at[0].add(-1)
        self.is_leaf = ~self.child_counts.astype(bool)
        self.is_root = jnp.zeros((self.size,), dtype=bool)
        self.is_root = self.is_root.at[0].set(True)
        self.is_inner = ~self.is_leaf & ~self.is_root

        pbuckets = jnp.copy(self.parents)
        pbuckets_ref = []
        for lmin,lmax in self.levels:
            pvals = pbuckets[lmin:lmax]
            pbuckets_refi, pbucket = jnp.unique(pvals, return_inverse=True)
            pbuckets = pbuckets.at[lmin:lmax].set(pbucket)
            pbuckets_ref.append(pbuckets_refi)
        self.pbuckets = pbuckets
        self.pbuckets_ref = pbuckets_ref

        #self.child_counts = jnp.array(child_counts)
        self.depth = len(self.levels)-1
        self.leaf_limits = self.levels[-1]

        self.coloring = jnp.zeros_like(self.parents, dtype=bool)
        for lmin, lmax in self.levels[::2]:
            self.coloring = self.coloring.at[lmin:lmax].set(True)

        if precompute_child_gathers:
            uniq = jnp.unique(self.child_counts)
            if len(uniq) == 2:
                null, base = uniq
                assert null == 0
                self.base = base
                #self.gather_child_idx = jnp.zeros((self.size, self.base), dtype=int)
                children = []
                for i,node in enumerate(self.iter_topology_bfs()):
                    if not node.children: children.append(int(base)*[0])
                    else: children.append([c.id for c in node.children])
                self.gather_child_idx = jnp.array(children) 
            else:
                raise ValueError("Only trees with the same number of children are supported")
            # 2 unique values implies that the tree has a "nice" structure



    def _calculate_levels(self):
        queue = deque()
        buffer_queue = deque([self.topology_root])
        while queue or buffer_queue:
            if not queue: # if queue is empty, flush the buffer and yield a level
                queue = buffer_queue
                yield (buffer_queue[0].id, buffer_queue[-1].id+1) # to not pass the reference
                buffer_queue = deque()

            if children := queue.popleft().children:
                buffer_queue.extend(children)


    def __len__(self) -> int:
        return self.size

    def add_property(self, name, shape, dtype = None, initializer = None, key = None):
        self.data[name] = jnp.empty((self.size, *shape), dtype=dtype)
        self.masks[name] = jnp.zeros((self.size,), dtype=bool)

    def iter_topology_bfs(self) -> Iterator[TopologyNode]:
        """
        Iterate over all of the nodes in a breadth first manner

        """
        queue = deque([self.topology_root])

        while queue:
            current = queue.popleft()
            if current.children:
                queue.extend(current.children)
            yield current
    
    def __repr__(self):
        """
        Return a string representation of the tree
        """
        return f"HypTree(size={self.size}, levels={len(self.levels)}, leaves={jnp.sum(self.is_leaf)}, inner nodes={jnp.sum(self.is_inner)})"

    def __str__(self):
        return self.__repr__()

    # Extra transversel methods for the tree
    def iter_topology_dfs(self) -> Iterator[TopologyNode]:
        """
        Iterate over all of the nodes in a depth-first manner.

        """
        stack = deque([self.topology_root])

        while stack:
            current = stack.pop()
            if current.children:
                stack.extend(current.children)
            yield current
    
    def iter_topology_leaves_dfs(self) -> Iterator[TopologyNode]:
        """
        Iterate over all of the leaves in the tree, in a depth-first manner.

        """

        queue = deque([self.topology_root])

        while queue:
            current = queue.popleft()
            if current.children:
                queue.extend(current.children)
            else:
                yield current
    
    def iter_topology_leaves_dfs(self) -> Iterator[TopologyNode]:
        """
        Iterates over the leaves in the tree using depth-first search.
        Note that this is not the same as iter_topology_leaves_dfs, as this method is bfs 
        """
        stack = deque([self.topology_root])

        while stack:
            current = stack.pop()
            if current.children:
                stack.extend(reversed(current.children))
            else:
                yield current

    def iter_topology_levels(self) -> Iterator[List[TopologyNode]]:
        """
        Iterate over each level in the tree

        """
        queue = deque()
        buffer_queue = deque([self.topology_root])
        while queue or buffer_queue:
            if not queue: # if queue is empty, flush the buffer and yield a level
                queue = buffer_queue
                yield list(buffer_queue) # to not pass the reference
                buffer_queue = deque()

            if children := queue.popleft().children:
                buffer_queue.extend(children)   

    # Extra transversel methods for the tree
    def iter_topology_dfs(self) -> Iterator[TopologyNode]:
        """
        Iterate over all of the nodes in a depth-first manner.

        """
        stack = deque([self.topology_root])

        while stack:
            current = stack.pop()
            if current.children:
                stack.extend(current.children)
            yield current
    
    def iter_topology_leaves_dfs(self) -> Iterator[TopologyNode]:
        """
        Iterate over all of the leaves in the tree, in a depth-first manner.

        """

        queue = deque([self.topology_root])

        while queue:
            current = queue.popleft()
            if current.children:
                queue.extend(current.children)
            else:
                yield current
    
    def iter_topology_leaves_dfs(self) -> Iterator[TopologyNode]:
        """
        Iterates over the leaves in the tree using depth-first search.
        Note that this is not the same as iter_topology_leaves_dfs, as this method is bfs 
        """
        stack = deque([self.topology_root])

        while stack:
            current = stack.pop()
            if current.children:
                stack.extend(reversed(current.children))
            else:
                yield current

    def iter_topology_levels(self) -> Iterator[List[TopologyNode]]:
        """
        Iterate over each level in the tree

        """
        queue = deque()
        buffer_queue = deque([self.topology_root])
        while queue or buffer_queue:
            if not queue: # if queue is empty, flush the buffer and yield a level
                queue = buffer_queue
                yield list(buffer_queue) # to not pass the reference
                buffer_queue = deque()

            if children := queue.popleft().children:
                buffer_queue.extend(children)   


class FastBiTree(HypTree):
    """
        Mostly for simulation purposes, creates a bifurcating tree
        without the topology, making instantiation fast
    """
    def __init__(self, depth) -> None:
        self.size = (2<<depth)-1
        self.data = {}
        self.masks = {}

        self.levels = [(0,1)]+[((2<<i)-1, (2<<(i+1))-1) for i in range(depth)]
        self.parents = jnp.concatenate([jnp.array([0]),jnp.arange(self.size-1)//2])
        self.depth = len(self.levels)-1
        self.leaves = self.levels[-1]

## would love to get this work but https://github.com/google/jax/issues/4269 makes 
## me believe this is probably a non JAX approach
class TreeField:
    def __init__(self, size, shape) -> None:
        raise NotImplementedError()
        self.shape = shape
        self.mask = jnp.zeros((size,), dtype=bool)
        self.data = jnp.empty((size, *shape))
    def __repr__(self) -> str:
        return str(self.data)


from jax import numpy as jnp
from jax.random import PRNGKey
from pytest import fixture
from hyperiax.tree.topology import symmetric_topology
from hyperiax.tree import HypTree
import jax



@fixture
def small_tree():
    return HypTree(symmetric_topology(5,2))


@fixture
def noise_tree():
    key = PRNGKey(0)
    t = HypTree(symmetric_topology(5,2))
    t.add_property('noise', (2,))
    t.data['noise'] = jax.random.normal(key, shape=t.data['noise'].shape)
    return t


